# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`LPAR` (Logical Partition) is a subset of the hardware resources of a
:term:`CPC` in classic mode (or ensemble mode), virtualized as a separate
computer.

LPARs cannot be created or deleted by the user; they can only be listed.

LPAR resources are contained in CPC resources.

LPAR resources only exist in CPCs that are in classic mode (or ensemble mode).
CPCs in DPM mode have :term:`Partition` resources, instead.
"""


import time
import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._exceptions import StatusTimeout
from ._constants import HMC_LOGGER_NAME
from ._logging import get_logger, logged_api_call
from ._utils import RC_LOGICAL_PARTITION, make_query_str, \
    warn_deprecated_parameter, datetime_from_timestamp, timestamp_from_datetime

__all__ = ['LparManager', 'Lpar']

HMC_LOGGER = get_logger(HMC_LOGGER_NAME)


class LparManager(BaseManager):
    """
    Manager providing access to the :term:`LPARs <LPAR>` in a particular
    :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object (in DPM mode):

    * :attr:`~zhmcclient.Cpc.lpars`

    HMC/SE version requirements: None
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
        ]

        super().__init__(
            resource_class=Lpar,
            class_name=RC_LOGICAL_PARTITION,
            session=cpc.manager.session,
            parent=cpc,
            base_uri='/api/logical-partitions',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props,
            supports_properties=True)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the LPARs in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any LPAR to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Lpar` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'logical-partitions'
        list_uri = f'{self.cpc.uri}/logical-partitions'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)


class Lpar(BaseResource):
    """
    Representation of an :term:`LPAR`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.LparManager`).

    HMC/SE version requirements: None
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.LparManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, LparManager), (
            f"Lpar init: Expected manager type {LparManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call(blanked_properties=['ssc-master-pw', 'zaware-master-pw'],
                     properties_pos=1)
    def update_properties(self, properties):
        """
        Update writeable properties of this LPAR.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Change Object Definition" task.
        * Since HMC 2.14.1: If the "next-activation-profile-name" property is to
          be updated, task permission for the "Change Object Options" task or
          the "Customize/Delete Activation Profiles" task.
        * Before HMC 2.15.0: For an LPAR whose activation-mode is "zaware", task
          permission for the "Firmware Details" task.
        * Since HMC 2.15.0: If any of the "ssc-*" or "zaware-*" properties is to
          be updated, task permission for the "Firmware Details" task.
        * Since HMC 2.15.0: If any of the numbers of allocated or reserved cores
          is to be updated, task permission for the "Logical Processor Add"
          task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Logical Partition object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        # Attempts to change the 'name' property will be rejected by the HMC,
        # so we don't need to update the name-to-URI cache.
        assert self.manager._name_prop not in properties
        self.update_properties_local(copy.deepcopy(properties))

    @logged_api_call
    def activate(self, wait_for_completion=True,
                 operation_timeout=None, status_timeout=None,
                 allow_status_exceptions=False, activation_profile_name=None,
                 force=False):
        """
        Activate (start) this LPAR, using the HMC operation "Activate Logical
        Partition".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it may take a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "not-operating" (which indicates that the LPAR is active but
        no operating system is running), or "operating", or if
        `allow_status_exceptions` was set additionally in the state
        "exceptions".

        The following approach is used to determine the desired state to wait
        for if `wait_for_completion=True`:

        - if the 'operating-mode' property of the image profile is 'ssc' or
          'zaware', the desired state is "operating".
        - if the profile specified in `activation_profile_name` is not the
          LPAR's image profile, it is assumed to be a load profile and
          the desired state is "operating".
        - if the 'load-at-activation' property of the image profile is True,
          the desired state is "operating".
        - else, the desired state is "not-operating".

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Activate" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "not-operating" or "operating" (or in addition
              "exceptions", if `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

          activation_profile_name (:term:`string`):
            Name of the load or image activation profile to be used instead
            of the one specified in the `next-activation-profile-name` property
            of the LPAR, or `None`.

            If this parameter specifies an image activation profile, its name
            must match the LPAR name. For non-SSC partitions, the image
            profile's `load-at-activation` property determines whether the
            activation is followed by a load of the control program using the
            load-related parameters from the image profile. SSC partitions are
            always auto-loaded (regardless of the `load-at-activation`
            property).

            If this parameter specifies a load activation profile, the
            activation uses the image profile with the same name as the LPAR.
            The activation is always followed by a load of the control program
            (regardless of the image profile's `load-at-activation` property)
            using the parameters from the load profile.

            If this parameter is `None`, the `next-activation-profile-name`
            property of the LPAR will be used. That property can again specify
            an image profile or a load profile which are treated as described
            above. If that property is `None`, the image profile with the same
            name as the LPAR is used and is treated as described above.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

            TBD: What will happen with the LPAR in that case (deactivated then
            activated? nothing?)

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        if activation_profile_name:
            body['activation-profile-name'] = activation_profile_name
        if force:
            body['force'] = force
        result = self.manager.session.post(
            self.uri + '/operations/activate', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            # If an automatic load is performed, the LPAR status will first go
            # to 'not-operating' and then later to 'operating'. So we cannot
            # just wait for any of those two, but need to have an understanding
            # whether we expect auto-load.
            image_profile_mgr = self.manager.parent.image_activation_profiles
            image_profile = image_profile_mgr.find(name=self.name)
            auto_load = image_profile.get_property('load-at-activation')
            # Note that the LPAR 'activation-mode' property is 'not-set' while
            # the LPAR is inactive, so we need to look at the image profile
            # to determine the mode.
            op_mode = image_profile.get_property('operating-mode')
            load_profile_specified = activation_profile_name is not None and \
                activation_profile_name != self.name
            mode_load = op_mode in ('ssc', 'zaware')
            if auto_load or load_profile_specified or mode_load:
                statuses = ["operating"]
            else:
                statuses = ["not-operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def deactivate(self, wait_for_completion=True,
                   operation_timeout=None, status_timeout=None,
                   allow_status_exceptions=False, force=False):
        """
        De-activate (stop) this LPAR, using the HMC operation "Deactivate
        Logical Partition".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it may take a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "not-activated", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Deactivate" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "not-activated" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

            TBD: What will happen with the LPAR in that case (deactivated then
            activated? nothing?)

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        if force:
            body['force'] = force
        result = self.manager.session.post(
            self.uri + '/operations/deactivate', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["not-activated"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def scsi_load(self, load_address, wwpn, lun, load_parameter=None,
                  disk_partition_id=None,
                  operating_system_specific_load_parameters=None,
                  boot_record_logical_block_address=None, force=False,
                  wait_for_completion=True, operation_timeout=None,
                  status_timeout=None, allow_status_exceptions=False,
                  secure_boot=False, os_ipl_token=None, clear_indicator=True):
        # pylint: disable=invalid-name
        """
        Load (boot) this LPAR from a designated SCSI device, using the
        HMC operation "SCSI Load".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "SCSI Load" task.

        Parameters:

          load_address (:term:`string`):
            Device number of the boot device.

          wwpn (:term:`string`):
            Worldwide port name (WWPN) of the target SCSI device to be
            used for this operation, in hexadecimal.

          lun (:term:`string`):
            Hexadecimal logical unit number (LUN) to be used for the
            SCSI load.

          load_parameter (:term:`string`):
            Optional load control string.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          disk_partition_id (:term:`integer`):
            Optional disk-partition-id (also called the boot program
            selector) to be used for the SCSI load.
            If `None`, it is not passed to the HMC, and the HMC default
            of 0 will be used.

          operating_system_specific_load_parameters (:term:`string`):
            Optional operating system specific load parameters to be
            used for the SCSI load.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          boot_record_logical_block_address (:term:`string`):
            Optional hexadecimal boot record logical block address to
            be used for the SCSI load.
            If `None`, it is not passed to the HMC, and the HMC default
            of "0" will be used.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

          secure_boot (bool):
            Boolean controlling whether the system checks the software
            signature of what is loaded against what the distributor signed it
            with.
            If `False` or `None`, it is not passed to the HMC, and the
            HMC default of `False` will be used.
            Requires the LPAR to be on a z15 or later.

          os_ipl_token (:term:`string`):
            Optional hexadecimal value to be used for the SCSI load.
            If `None`, it is not passed to the HMC.

          clear_indicator (bool):
            Optional boolean controlling whether the memory should be
            cleared before performing the load or not cleared.
            If `True` or `None`, it is not passed to the HMC, and the HMC
            default of `True` will be used if the LPAR is on a z14 with
            SE version 2.14.1 or higher.
            Requires the LPAR to be on a z14 with SE version 2.14.1 or higher.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        body['load-address'] = load_address
        body['world-wide-port-name'] = wwpn
        body['logical-unit-number'] = lun
        if load_parameter:
            body['load-parameter'] = load_parameter
        if disk_partition_id is not None:
            body['disk-partition-id'] = disk_partition_id
        if operating_system_specific_load_parameters:
            body['operating-system-specific-load-parameters'] = \
                operating_system_specific_load_parameters
        if boot_record_logical_block_address is not None:
            body['boot-record-logical-block-address'] = \
                boot_record_logical_block_address
        if os_ipl_token is not None:
            body['os-ipl-token'] = os_ipl_token
        if clear_indicator not in (True, None):
            # Note: Requires SE >= 2.14.1, but caller needs to control this
            body['clear-indicator'] = clear_indicator
        if force:
            body['force'] = force
        if secure_boot:
            # Note: Requires SE >= 2.15, but caller needs to control this
            body['secure-boot'] = secure_boot
        result = self.manager.session.post(
            self.uri + '/operations/scsi-load', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def scsi_dump(self, load_address, wwpn, lun, load_parameter=None,
                  disk_partition_id=None,
                  operating_system_specific_load_parameters=None,
                  boot_record_logical_block_address=None, os_ipl_token=None,
                  wait_for_completion=True, operation_timeout=None,
                  status_timeout=None, allow_status_exceptions=False,
                  force=False, secure_boot=False):
        # pylint: disable=invalid-name
        """
        Load a standalone dump program from a designated SCSI device
        in this LPAR, using the HMC operation "SCSI Dump".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "SCSI Dump" task.

        Parameters:

          load_address (:term:`string`):
            Device number of the boot device.

          wwpn (:term:`string`):
            Worldwide port name (WWPN) of the target SCSI device to be
            used for this operation, in hexadecimal.

          lun (:term:`string`):
            Hexadecimal logical unit number (LUN) to be used for the
            SCSI dump.

          load_parameter (:term:`string`):
            Optional load control string.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          disk_partition_id (:term:`integer`):
            Optional disk-partition-id (also called the boot program
            selector) to be used for the SCSI dump.
            If `None`, it is not passed to the HMC, and the HMC default
            of 0 will be used.

          operating_system_specific_load_parameters (:term:`string`):
            Optional operating system specific load parameters to be
            used for the SCSI dump.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          boot_record_logical_block_address (:term:`string`):
            Optional hexadecimal boot record logical block address to
            be used for the SCSI dump.
            If `None`, it is not passed to the HMC, and the HMC default
            of "0" will be used.

          os_ipl_token (:term:`string`):
            Optional hexadecimal value to be used for the SCSI dump.
            If `None`, it is not passed to the HMC.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

          secure_boot (bool):
            Boolean controlling whether the system checks the software
            signature of what is loaded against what the distributor signed it
            with.
            If `False` or `None`, it is not passed to the HMC, and the
            HMC default of `False` will be used.
            Requires the LPAR to be on a z15 or later.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        body['load-address'] = load_address
        body['world-wide-port-name'] = wwpn
        body['logical-unit-number'] = lun
        if load_parameter:
            body['load-parameter'] = load_parameter
        if disk_partition_id is not None:
            body['disk-partition-id'] = disk_partition_id
        if operating_system_specific_load_parameters:
            body['operating-system-specific-load-parameters'] = \
                operating_system_specific_load_parameters
        if boot_record_logical_block_address is not None:
            body['boot-record-logical-block-address'] = \
                boot_record_logical_block_address
        if os_ipl_token is not None:
            body['os-ipl-token'] = os_ipl_token
        if force:
            body['force'] = force
        if secure_boot:
            # Note: Requires SE >= 2.15, but caller needs to control this
            body['secure-boot'] = secure_boot
        result = self.manager.session.post(
            self.uri + '/operations/scsi-dump', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def nvme_load(self, load_address, load_parameter=None, secure_boot=False,
                  clear_indicator=True, disk_partition_id=None,
                  operating_system_specific_load_parameters=None,
                  boot_record_logical_block_address=None, force=False,
                  wait_for_completion=True, operation_timeout=None,
                  status_timeout=None, allow_status_exceptions=False):
        # pylint: disable=invalid-name
        """
        Load (boot) this LPAR from a designated NVMe device, using the
        HMC operation "NVMe Load".

        This operation requires z15 or later.

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements:

        * SE version >= 2.15.0

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Task permission for the "Load" task.

        Parameters:

          load_address (:term:`string`):
            Device number of the boot device.

          load_parameter (:term:`string`):
            Optional load control string.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          secure_boot (bool):
            Boolean controlling whether the system checks the software
            signature of what is loaded against what the distributor signed it
            with.
            If `False` or `None`, it is not passed to the HMC, and the
            HMC default of `False` will be used.
            Requires the LPAR to be on a z15 or later.

          clear_indicator (bool):
            Optional boolean controlling whether the memory should be
            cleared before performing the load or not cleared.
            If `True` or `None`, it is not passed to the HMC, and the HMC
            default of `True` will be used if the LPAR is on a z14 with
            SE version 2.14.1 or higher.
            Requires the LPAR to be on a z14 with SE version 2.14.1 or higher.

          disk_partition_id (:term:`integer`):
            Optional disk-partition-id (also called the boot program
            selector) to be used for the NVMe Load.
            If `None`, it is not passed to the HMC, and the HMC default
            of 0 will be used.

          operating_system_specific_load_parameters (:term:`string`):
            Optional operating system specific load parameters to be
            used for the NVMe Load.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          boot_record_logical_block_address (:term:`string`):
            Optional hexadecimal boot record logical block address to
            be used for the NVMe Load.
            If `None`, it is not passed to the HMC, and the HMC default
            of "0" will be used.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        body['load-address'] = load_address
        if load_parameter:
            body['load-parameter'] = load_parameter
        if disk_partition_id is not None:
            body['disk-partition-id'] = disk_partition_id
        if operating_system_specific_load_parameters:
            body['operating-system-specific-load-parameters'] = \
                operating_system_specific_load_parameters
        if boot_record_logical_block_address is not None:
            body['boot-record-logical-block-address'] = \
                boot_record_logical_block_address
        if clear_indicator not in (True, None):
            # Note: Requires SE >= 2.14.1, but caller needs to control this
            body['clear-indicator'] = clear_indicator
        if force:
            body['force'] = force
        if secure_boot:
            # Note: Requires SE >= 2.15, but caller needs to control this
            body['secure-boot'] = secure_boot
        result = self.manager.session.post(
            self.uri + '/operations/nvme-load', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def nvme_dump(self, load_address, load_parameter=None,
                  secure_boot=False, disk_partition_id=None,
                  operating_system_specific_load_parameters=None,
                  boot_record_logical_block_address=None, force=False,
                  wait_for_completion=True, operation_timeout=None,
                  status_timeout=None, allow_status_exceptions=False):
        # pylint: disable=invalid-name
        """
        Load a standalone dump program from a designated NVMe device
        in this LPAR, using the HMC operation "NVMe Dump".

        This operation requires z15 or later.

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements:

        * SE version >= 2.15.0

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Task permission for the "NVMe Dump" task.

        Parameters:

          load_address (:term:`string`):
            Device number of the boot device.

          load_parameter (:term:`string`):
            Optional load control string.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          secure_boot (bool):
            Boolean controlling whether the system checks the software
            signature of what is loaded against what the distributor signed it
            with.
            If `False` or `None`, it is not passed to the HMC, and the
            HMC default of `False` will be used.
            Requires the LPAR to be on a z15 or later.

          disk_partition_id (:term:`integer`):
            Optional disk-partition-id (also called the boot program
            selector) to be used for the NVMe dump.
            If `None`, it is not passed to the HMC, and the HMC default
            of 0 will be used.

          operating_system_specific_load_parameters (:term:`string`):
            Optional operating system specific load parameters to be
            used for the NVMe dump.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          boot_record_logical_block_address (:term:`string`):
            Optional hexadecimal boot record logical block address to
            be used for the NVMe dump.
            If `None`, it is not passed to the HMC, and the HMC default
            of "0" will be used.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        body['load-address'] = load_address
        if load_parameter:
            body['load-parameter'] = load_parameter
        if disk_partition_id is not None:
            body['disk-partition-id'] = disk_partition_id
        if operating_system_specific_load_parameters:
            body['operating-system-specific-load-parameters'] = \
                operating_system_specific_load_parameters
        if boot_record_logical_block_address is not None:
            body['boot-record-logical-block-address'] = \
                boot_record_logical_block_address
        if force:
            body['force'] = force
        if secure_boot:
            # Note: Requires SE >= 2.15, but caller needs to control this
            body['secure-boot'] = secure_boot
        result = self.manager.session.post(
            self.uri + '/operations/nvme-dump', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def load(self, load_address=None, load_parameter=None,
             clear_indicator=True, store_status_indicator=False,
             wait_for_completion=True, operation_timeout=None,
             status_timeout=None, allow_status_exceptions=False,
             force=False):
        """
        Load (boot) this LPAR from a load address (boot device), using the HMC
        operation "Load Logical Partition".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Load" task.

        Parameters:

          load_address (:term:`string`): Device number of the boot device.
            Up to z13, this parameter is required.
            Starting with z14, this parameter is optional and defaults to the
            load address specified in the 'last-used-load-address' property of
            the Lpar.

          load_parameter (:term:`string`): Optional load control string.
            If empty string or `None`, it is not passed to the HMC, and the
            HMC default of an empty string will be used.

          clear_indicator (bool):
            Optional boolean controlling whether the memory should be
            cleared before performing the load or not cleared. The
            default value is `True`.

          store_status_indicator (bool):
            Optional boolean controlling whether the status should be
            stored before performing the Load. The default value is `False`.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status.

            TBD: What will happen with the LPAR in that case (deactivated then
            activated? nothing?)

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {}
        if load_address:
            body['load-address'] = load_address
        if load_parameter:
            body['load-parameter'] = load_parameter
        if force:
            body['force'] = force
        if not clear_indicator:
            body['clear-indicator'] = clear_indicator
        if store_status_indicator:
            body['store-status-indicator'] = store_status_indicator
        result = self.manager.session.post(
            self.uri + '/operations/load', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def load_from_ftp(
            self, host, username, password, load_file, protocol='ftp',
            wait_for_completion=True, operation_timeout=None,
            status_timeout=None, allow_status_exceptions=False):
        """
        Load (boot) this LPAR from an FTP server, using the HMC operation
        "Load Logical Partition from FTP".

        This operation is not permitted for an LPAR whose 'activation-mode'
        property is "zaware" or "ssc".

        This HMC operation has deferred status behavior: If the asynchronous
        job on the HMC is complete, it takes a few seconds until the LPAR
        status has reached the desired value. If `wait_for_completion=True`,
        this method repeatedly checks the status of the LPAR after the HMC
        operation has completed, and waits until the status is in the desired
        state "operating", or if `allow_status_exceptions` was
        set additionally in the state "exceptions".

        HMC/SE version requirements:

        * SE version >= 2.16.0

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Load from Removable Media or Server" task.

        Parameters:

          host (string): Host name or IP address of the FTP server.

          username (string): User name for the account on the FTP server.

          password (string): Password that is associated with the user name on
            the FTP server.

          load_file (string): Path name of the file to be read from the FTP
            server and loaded into the LPAR.

          protocol (string): Network protocol for transferring files. Must be
            one of:

              * "ftp" - File Transfer Protocol
              * "ftps" - FTP Secure
              * "sftp" - SSH File Transfer Protocol

            Default: "ftp"

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation, and for the status
              becoming "operating" (or in addition "exceptions", if
              `allow_status_exceptions` was set.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached the desired status, after the HMC operation has completed.
            The special value 0 means that no timeout is set. `None` means that
            the default status timeout of the session is used.
            If the timeout expires when `wait_for_completion=True`, a
            :exc:`~zhmcclient.StatusTimeout` is raised.

          allow_status_exceptions (bool):
            Boolean controlling whether LPAR status "exceptions" is considered
            an additional acceptable end status when `wait_for_completion` is
            set.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        body = {
            'host-name': host,
            'user-name': username,
            'password': password,
            'file-path': load_file,
            'protocol': protocol,
        }
        result = self.manager.session.post(
            self.uri + '/operations/load-from-ftp', body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        if wait_for_completion:
            statuses = ["operating"]
            if allow_status_exceptions:
                statuses.append("exceptions")
            self.wait_for_status(statuses, status_timeout)
        return result

    @logged_api_call
    def stop(self, wait_for_completion=True, operation_timeout=None,
             status_timeout=None, allow_status_exceptions=False):
        """
        Stop this LPAR, using the HMC operation "Stop Logical
        Partition". The stop operation stops the processors from
        processing instructions.

        This operation is not permitted for an LPAR whose 'activation-mode'
        property is "zaware" or "ssc".

        In order to succeed, the 'status' property of the LPAR must have one of
        the following values:

        * "not-operating"
        * "operating"
        * "exceptions"

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Stop" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          allow_status_exceptions (bool):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        warn_deprecated_parameter(
            Lpar, Lpar.stop, 'status_timeout', status_timeout, None)
        warn_deprecated_parameter(
            Lpar, Lpar.stop, 'allow_status_exceptions',
            allow_status_exceptions, False)
        body = None
        result = self.manager.session.post(
            self.uri + '/operations/stop', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def start(self, wait_for_completion=True, operation_timeout=None,
              status_timeout=None, allow_status_exceptions=False):
        """
        Start this LPAR, using the HMC operation "Start Logical
        Partition". The start operation starts the processors to process
        instructions.

        This operation is not permitted for an LPAR whose 'activation-mode'
        property is "zaware" or "ssc".

        In order to succeed, the 'status' property of the LPAR must have one of
        the following values:

        * "not-operating"
        * "operating"
        * "exceptions"

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Start" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          allow_status_exceptions (bool):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        warn_deprecated_parameter(
            Lpar, Lpar.start, 'status_timeout', status_timeout, None)
        warn_deprecated_parameter(
            Lpar, Lpar.start, 'allow_status_exceptions',
            allow_status_exceptions, False)
        body = None
        result = self.manager.session.post(
            self.uri + '/operations/start', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def reset_clear(self, force=False, wait_for_completion=True,
                    operation_timeout=None, status_timeout=None,
                    allow_status_exceptions=False, os_ipl_token=None):
        """
        Reset this LPAR and clears its memory.

        This includes clearing its pending interruptions, resetting its channel
        subsystem and resetting its processors, and clearing its memory, using
        the HMC operation "Reset Clear".

        In order to succeed, the 'status' property of the LPAR must have one of
        the following values:

        * "not-operating"
        * "operating" - this requires setting the "force" flag
        * "exceptions" - this requires setting the "force" flag

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Reset Clear" task.

        Parameters:

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status. The default is `False`.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          allow_status_exceptions (bool):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          os_ipl_token (:term:`string`):
            Applicable only to z/OS, this parameter requests that this
            operation only be performed if the provided value matches the
            current value of the 'os-ipl-token' property of the LPAR.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        warn_deprecated_parameter(
            Lpar, Lpar.reset_clear, 'status_timeout', status_timeout, None)
        warn_deprecated_parameter(
            Lpar, Lpar.reset_clear, 'allow_status_exceptions',
            allow_status_exceptions, False)
        body = {}
        if force:
            body['force'] = force
        if os_ipl_token:
            body['os-ipl-token'] = os_ipl_token
        result = self.manager.session.post(
            self.uri + '/operations/reset-clear', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def reset_normal(self, force=False, wait_for_completion=True,
                     operation_timeout=None, status_timeout=None,
                     allow_status_exceptions=False, os_ipl_token=None):
        """
        Reset this LPAR without clearing its memory.

        This includes clearing its pending interruptions, resetting its channel
        subsystem and resetting its processors, using the HMC operation
        "Reset Normal".

        In order to succeed, the 'status' property of the LPAR must have one of
        the following values:

        * "not-operating"
        * "operating" - this requires setting the "force" flag
        * "exceptions" - this requires setting the "force" flag

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "Reset Clear" task.

        Parameters:

          force (bool):
            Boolean controlling whether this operation is permitted when the
            LPAR is in the "operating" status. The default is `False`.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          allow_status_exceptions (bool):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          os_ipl_token (:term:`string`):
            Applicable only to z/OS, this parameter requests that this
            operation only be performed if the provided value matches the
            current value of the 'os-ipl-token' property of the LPAR.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        warn_deprecated_parameter(
            Lpar, Lpar.reset_normal, 'status_timeout', status_timeout, None)
        warn_deprecated_parameter(
            Lpar, Lpar.reset_normal, 'allow_status_exceptions',
            allow_status_exceptions, False)
        body = {}
        if force:
            body['force'] = force
        if os_ipl_token:
            body['os-ipl-token'] = os_ipl_token
        result = self.manager.session.post(
            self.uri + '/operations/reset-normal', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def open_os_message_channel(self, include_refresh_messages=True):
        """
        Open a JMS message channel to this LPAR's operating system, returning
        the string "topic" representing the message channel.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Operating System Messages" task at least
          in view-only mode.

        Parameters:

          include_refresh_messages (bool):
            Boolean controlling whether refresh operating systems messages
            should be sent, as follows:

            * If `True`, refresh messages will be recieved when the user
              connects to the topic. The default.

            * If `False`, refresh messages will not be recieved when the user
              connects to the topic.

        Returns:

          :term:`string`:

            Returns a string representing the os-message-notification JMS
            topic. The user can connect to this topic to start the flow of
            operating system messages.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'include-refresh-messages': include_refresh_messages}
        result = self.manager.session.post(
            self.uri + '/operations/open-os-message-channel',
            resource=self, body=body)
        return result['topic-name']

    @logged_api_call
    def send_os_command(self, os_command_text, is_priority=False):
        """
        Send a command to the operating system running in this LPAR.

        HMC/SE version requirements:

        * SE version >= 2.13.1

        Authorization requirements:

        * Object-access permission to this Partition.
        * Task permission to the "Operating System Messages" task in
          modification mode.

        Parameters:

          os_command_text (string): The text of the operating system command.

          is_priority (bool):
            Boolean controlling whether this is a priority operating system
            command, as follows:

            * If `True`, this message is treated as a priority operating
              system command.

            * If `False`, this message is not treated as a priority
              operating system command. The default.

        Returns:

          None

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'is-priority': is_priority,
                'operating-system-command-text': os_command_text}
        self.manager.session.post(
            self.uri + '/operations/send-os-cmd', resource=self, body=body)

    @logged_api_call
    def list_os_messages(
            self, begin=None, end=None, is_held=None, is_priority=None,
            max_messages=0):
        """
        List all currently available operating system messages for this
        LPAR.

        Only a certain amount of OS message data from each LPAR is
        preserved by the HMC for retrieval by this operation. If the OS
        produces more than that amount, the oldest non-held, non-priority
        OS messages are no longer available. A gap in the sequence numbers
        indicates a loss of messages. A loss may be due to that space
        limitation, or it may be due to the deletion of messages by a console
        user or the OS.

        HMC/SE version requirements:

        * SE version >= 2.14.0

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Task permission to the "Operating System Messages" task (optionally
          in view-only mode).

        Parameters:

          begin (integer): A message sequence number to limit returned
            messages. OS messages with a sequence number less than this are
            omitted from the results. If `None`, no such filtering is
            performed.

          end (integer): A message sequence number to limit returned
            messages. OS messages with a sequence number greater than this are
            omitted from the results. If `None`, no such filtering is
            performed.

          is_held(bool): Limit the returned messages to only held (if `True`)
            or only non-held (if `False`) messages. If `None`, no such filtering
            is performed.

          is_priority(bool): Limit the returned messages to only priority (if
            `True`) or non-priority (if `False`) messages. If `None`, no such
            filtering is performed.

          max_messages(int): Limits the returned messages to the specified
            maximum number, starting from the begin of the sequence numbers
            in the result that would otherwise be returned.
            If 0, no such filtering is performed.

        Returns:

          list of dict: List of OS messages, where each OS message is a dict
          with the items defined for the "os-message-info" data structure
          in the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms = []
        if begin is not None:
            query_parms.append(f'begin-sequence-number={begin}')
        if end is not None:
            query_parms.append(f'end-sequence-number={end}')
        if is_held is not None:
            query_parms.append(f'is-held={str(is_held).lower()}')
        if is_priority is not None:
            query_parms.append(
                f'is-priority={str(is_priority).lower()}')
        if max_messages > 0:
            query_parms.append(f'max-messages={max_messages}')
        query_str = make_query_str(query_parms)
        result = self.manager.session.get(
            f'{self.uri}/operations/list-os-messages{query_str}',
            resource=self)
        return result

    @logged_api_call
    def psw_restart(self, wait_for_completion=True, operation_timeout=None,
                    status_timeout=None, allow_status_exceptions=False):
        """
        Restart this LPAR, using the HMC operation "PSW Restart".

        In order to succeed, the 'status' property of the LPAR must have one of
        the following values:

        * "not-operating"
        * "operating"
        * "exceptions"

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Before HMC API version 3.6 in an update to HMC 2.15.0: Object-access
          permission to the CPC of this LPAR.
        * Task permission for the "PSW Restart" task.

        Parameters:

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation, as follows:

            * If `True`, this method will wait for completion of the
              asynchronous job performing the operation.

            * If `False`, this method will return immediately once the HMC has
              accepted the request to perform the operation.

          operation_timeout (:term:`number`):
            Timeout in seconds, for waiting for completion of the asynchronous
            job performing the operation. The special value 0 means that no
            timeout is set. `None` means that the default async operation
            timeout of the session is used. If the timeout expires when
            `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised.

          status_timeout (:term:`number`):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

          allow_status_exceptions (bool):
            **Deprecated:** This property was used for handling deferred status
            behavior, which is not actually needed. Setting it to a non-default
            value will cause a :exc:`~py:exceptions.DeprecationWarning` to be
            issued.

        Returns:

          `None` or :class:`~zhmcclient.Job`:

            If `wait_for_completion` is `True`, returns `None`.

            If `wait_for_completion` is `False`, returns a
            :class:`~zhmcclient.Job` object representing the asynchronously
            executing job on the HMC.
            This job does not support cancellation.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the operation.
        """
        warn_deprecated_parameter(
            Lpar, Lpar.psw_restart, 'status_timeout', status_timeout, None)
        warn_deprecated_parameter(
            Lpar, Lpar.psw_restart, 'allow_status_exceptions',
            allow_status_exceptions, False)
        body = None
        result = self.manager.session.post(
            self.uri + '/operations/psw-restart', resource=self, body=body,
            wait_for_completion=wait_for_completion,
            operation_timeout=operation_timeout)
        return result

    @logged_api_call
    def wait_for_status(self, status, status_timeout=None):
        """
        Wait until the status of this LPAR has a desired value.

        HMC/SE version requirements: None

        Parameters:

          status (:term:`string` or iterable of :term:`string`):
            Desired LPAR status or set of status values to reach; one or more
            of the following values:

            * ``"not-activated"`` - The LPAR is not active.
            * ``"not-operating"`` - The LPAR is active but no operating system
              is running in the LPAR.
            * ``"operating"`` - The LPAR is active and an operating system is
              running in the LPAR.
            * ``"exceptions"`` - The LPAR or its CPC has one or more unusual
              conditions.

            Note that the description of LPAR status values in the
            :term:`HMC API` book (as of its version 2.13.1) is partly
            confusing.

          status_timeout (:term:`number`):
            Timeout in seconds, for waiting that the status of the LPAR has
            reached one of the desired status values. The special value 0 means
            that no timeout is set.
            `None` means that the default status timeout will be used.
            If the timeout expires , a :exc:`~zhmcclient.StatusTimeout` is
            raised.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.StatusTimeout`: The timeout expired while
            waiting for the desired LPAR status.
        """
        if status_timeout is None:
            status_timeout = \
                self.manager.session.retry_timeout_config.status_timeout
        if status_timeout > 0:
            end_time = time.time() + status_timeout
        if isinstance(status, (list, tuple)):
            statuses = status
        else:
            statuses = [status]
        HMC_LOGGER.debug("Waiting for LPAR %r to have status: %s "
                         "(timeout: %s sec)",
                         self.name, status, status_timeout)
        while True:

            # Fastest way to get actual status value:
            lpars = self.manager.cpc.lpars.list(
                filter_args={'name': self.name})
            assert len(lpars) == 1
            this_lpar = lpars[0]
            actual_status = this_lpar.get_property('status')

            if actual_status in statuses:
                return

            # pylint: disable=possibly-used-before-assignment
            if status_timeout > 0 and time.time() > end_time:
                raise StatusTimeout(
                    f"Waiting for LPAR {self.name} to reach status(es) "
                    f"'{statuses}' timed out after {status_timeout} s - "
                    f"current status is '{actual_status}'",
                    actual_status, statuses, status_timeout)

            time.sleep(1)  # Avoid hot spin loop

    @logged_api_call
    def assign_certificate(self, certificate):
        """
        Assigns a :term:`Certificate` to this LPAR.

        HMC/SE version requirements:

        * :ref`API feature` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be assigned. The certificate must not currently
            be assigned to this LPAR.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/assign-certificate', resource=self,
            body=body)

    @logged_api_call
    def unassign_certificate(self, certificate):
        """
        Unassign a :term:`Certificate` from this LPAR.

        HMC/SE version requirements:

        * :ref`API feature` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this LPAR.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be unassigned. The certificate must currently be
            assigned to this LPAR.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/unassign-certificate', resource=self,
            body=body)

    @logged_api_call
    def get_sustainability_data(
            self, range="last-week", resolution="one-hour",
            custom_range_start=None, custom_range_end=None):
        # pylint: disable=redefined-builtin
        """
        Get energy management related metrics for the LPAR on a specific
        historical time range. The metrics are returned as multiple data points
        covering the requested time range with the requested resolution.
        This method performs the "Get LPAR Historical Sustainability Data"
        HMC operation.

        HMC/SE version requirements:

        * :ref`API feature` "environmental-metrics"

        Authorization requirements:

        * Object-access permission to this LPAR
        * Task permission to the "Environmental Dashboard" task

        Parameters:

          range (:term:`string`):
            Time range for the requested data points, as follows:

            * "last-day" - Last 24 hours.
            * "last-week" - Last 7 days (default).
            * "last-month" - Last 30 days.
            * "last-three-months" - Last 90 days.
            * "last-six-months" - Last 180 days.
            * "last-year" - Last 365 days.
            * "custom" - From `custom_range_start` to `custom_range_end`.

          resolution (:term:`string`):
            Resolution for the requested data points. This is the time interval
            in between the data points. For systems where the
            "environmental-metrics" API feature is not available, the minimum
            resolution is "one-hour".

            The possible values are as follows:

            * "fifteen-minutes" - 15 minutes.
            * "one-hour" - 60 minutes (default).
            * "one-day" - 24 hours.
            * "one-week" - 7 days.
            * "one-month" - 30 days.

          custom_range_start (:class:`~py:datetime.datetime`):
            Start of custom time range. Timezone-naive values are interpreted
            using the local system time. Required if `range` is "custom".

          custom_range_end (:class:`~py:datetime.datetime`):
            End of custom time range. Timezone-naive values are interpreted
            using the local system time. Required if `range` is "custom".

        Returns:

          dict: A dictionary with items as described for the response body
          of the "Get LPAR Historical Sustainability Data" HMC operation.
          Timestamp fields are represented as timezone-aware
          :class:`~py:datetime.datetime` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        body = {
            'range': range,
            'resolution': resolution,
        }
        if range == "custom":
            body['custom-range-start'] = \
                timestamp_from_datetime(custom_range_start)
            body['custom-range-end'] = \
                timestamp_from_datetime(custom_range_end)
        result = self.manager.session.post(
            self.uri + '/operations/get-historical-sustainability-data',
            body=body)
        for field_array in result.values():
            for item in field_array:
                if 'timestamp' in item:
                    item['timestamp'] = \
                        datetime_from_timestamp(item['timestamp'])
        return result



# Copyright 2018,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :class:`~zhmcclient.StorageVolume` object represents an FCP or FICON (ECKD)
:term:`storage volume` that is defined in a :term:`storage group`.

Storage volume objects can be created, but what is created is the definition
of a storage volume in the HMC and CPC. This does not include the act of
actually allocating the volume on a storage subsystem. That is performed by a
storage administrator who :term:`fulfills <fulfillment>` the volumes.

In order to represent that, storage volume objects have a fulfillment state
that is available in their 'fulfillment-state' property.

When a storage group is attached to a :term:`partition`, the group's
storage volumes are attached to the partition and any :term:`virtual storage
resource` objects are instantiated.

Storage volumes are contained in :term:`storage groups <storage group>`.

You can create as many storage volumes as you want in a storage group.

Storage groups and storage volumes only can be defined in CPCs that are in
DPM mode and that have the "dpm-storage-management"
:ref:`firmware feature <firmware features>` enabled.
"""


import re
import copy
# from requests.utils import quote

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_STORAGE_VOLUME

__all__ = ['StorageVolumeManager', 'StorageVolume']


class StorageVolumeManager(BaseManager):
    """
    Manager providing access to the :term:`storage volumes <storage volume>`
    in a particular :term:`storage group`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.StorageGroup` object:

    * :attr:`~zhmcclient.StorageGroup.storage_volumes`

    HMC/SE version requirements:

    * :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, storage_group):
        # This function should not go into the docs.
        # Parameters:
        #   storage_group (:class:`~zhmcclient.StorageGroup`):
        #     Storage group defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
            'fulfillment-state',
            'maximum-size',
            'minimum-size',
            'usage',
        ]

        super().__init__(
            resource_class=StorageVolume,
            class_name=RC_STORAGE_VOLUME,
            session=storage_group.manager.session,
            parent=storage_group,
            base_uri=f'{storage_group.uri}/storage-volumes',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def storage_group(self):
        """
        :class:`~zhmcclient.StorageGroup`: :term:`Storage group` defining the
        scope for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the storage volumes in this storage group.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.

        Parameters:

          full_properties (bool):
            Controls that the full set of resource properties for each returned
            storage volume is being retrieved, vs. only the following short
            set: "element-uri", "name", "fulfillment-state", "size", and
            "usage".

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.StorageVolume` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'storage-volumes'
        list_uri = f'{self.storage_group.uri}/storage-volumes'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties, email_to_addresses=None,
               email_cc_addresses=None, email_insert=None):
        """
        Create a :term:`storage volume` in this storage group on the HMC, and
        optionally send emails to storage administrators requesting creation of
        the storage volume on the storage subsystem and setup of any resources
        related to the storage volume (e.g. LUN mask definition on the storage
        subsystem).

        This method performs the "Modify Storage Group Properties" operation,
        requesting creation of the volume.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this storage group.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): Initial property values for the new volume.

            Allowable properties are the fields defined in the
            "storage-volume-request-info" nested object described for
            operation "Modify Storage Group Properties" in the
            :term:`HMC API` book.
            The valid fields are those for the "create" operation. The
            `operation` field must not be provided; it is set automatically
            to the value "create".

            The properties provided in this parameter will be copied and then
            amended with the `operation="create"` field, and then used as a
            single array item for the `storage-volumes` field in the request
            body of the "Modify Storage Group Properties" operation.

            Note that for storage volumes, the HMC does auto-generate a value
            for the "name" property, but that auto-generated name is not unique
            within the parent storage group. If you depend on a unique name,
            you need to specify a "name" property accordingly.

          email_to_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be notified.
            If `None` or empty, no email will be sent.

          email_cc_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be copied
            on the notification email.
            If `None` or empty, nobody will be copied on the email.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

          email_insert (:term:`string`): Additional text to be inserted in the
            notification email.
            The text can include HTML formatting tags.
            If `None`, no additional text will be inserted.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

        Returns:

          StorageVolume:
            The resource object for the new storage volume.
            The object will have the following properties set:

            - 'element-uri' as returned by the HMC
            - 'element-id' as determined from the 'element-uri' property
            - 'class' and 'parent'
            - additional properties as specified in the input properties

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`

        Example::

            stovol1 = fcp_stogrp.storage_volumes.create(
                properties=dict(
                    name='vol1',
                    size=30,  # GiB
            ))
        """

        volreq_obj = copy.deepcopy(properties)
        volreq_obj['operation'] = 'create'
        body = {
            'storage-volumes': [volreq_obj],
        }
        if email_to_addresses:
            body['email-to-addresses'] = email_to_addresses
            if email_cc_addresses:
                body['email-cc-addresses'] = email_cc_addresses
            if email_insert:
                body['email-insert'] = email_insert
        else:
            if email_cc_addresses:
                raise ValueError(
                    "email_cc_addresses must not be specified if there is no "
                    f"email_to_addresses: {email_cc_addresses}")
            if email_insert:
                raise ValueError(
                    "email_insert must not be specified if there is no "
                    f"email_to_addresses: {email_insert}")

        result = self.session.post(
            self.storage_group.uri + '/operations/modify',
            body=body)
        uri = result['element-uris'][0]

        storage_volume = self.resource_object(uri, properties)

        # The name is not guaranteed to be unique, so we don't maintain
        # a name-to-uri cache for storage volumes.

        return storage_volume


class StorageVolume(BaseResource):
    """
    Representation of a :term:`storage volume`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.StorageVolumeManager`).

    HMC/SE version requirements:

    * :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.StorageVolumeManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, StorageVolumeManager), (
            "StorageVolume init: Expected manager type "
            f"{StorageVolumeManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @property
    def oid(self):
        """
        :term `unicode string`: The object ID of this storage volume.
        The object ID is unique within the parent storage group.

        Note that for storage volumes, the 'name' property is not unique and
        therefore cannot be used to identify a storage volume. Therefore,
        storage volumes provide easy access to the object ID, as a means to
        identify the storage volume in CLIs and other string-based tooling.
        """
        m = re.match(r'^/api/storage-groups/[^/]*/storage-volumes/([^/]*)$',
                     self.uri)
        oid = m.group(1)
        return oid

    @logged_api_call
    def delete(self, email_to_addresses=None, email_cc_addresses=None,
               email_insert=None):
        """
        Delete this storage volume on the HMC, and optionally send emails to
        storage administrators requesting deletion of the storage volume on the
        storage subsystem and cleanup of any resources related to the storage
        volume (e.g. LUN mask definitions on a storage subsystem).

        This method performs the "Modify Storage Group Properties" operation,
        requesting deletion of the volume.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the storage group owning this storage
          volume.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          email_to_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be notified.
            If `None` or empty, no email will be sent.

          email_cc_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be copied
            on the notification email.
            If `None` or empty, nobody will be copied on the email.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

          email_insert (:term:`string`): Additional text to be inserted in the
            notification email.
            The text can include HTML formatting tags.
            If `None`, no additional text will be inserted.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        volreq_obj = {
            'operation': 'delete',
            'element-uri': self.uri,
        }
        body = {
            'storage-volumes': [
                volreq_obj
            ],
        }
        if email_to_addresses:
            body['email-to-addresses'] = email_to_addresses
            if email_cc_addresses:
                body['email-cc-addresses'] = email_cc_addresses
            if email_insert:
                body['email-insert'] = email_insert
        else:
            if email_cc_addresses:
                raise ValueError(
                    "email_cc_addresses must not be specified if there is no "
                    f"email_to_addresses: {email_cc_addresses!r}")
            if email_insert:
                raise ValueError(
                    "email_insert must not be specified if there is no "
                    f"email_to_addresses: {email_insert}")

        self.manager.session.post(
            self.manager.storage_group.uri + '/operations/modify',
            resource=self, body=body)

        # pylint: disable=protected-access
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))

        parent_sv_uris = self.manager.parent.get_properties_local(
            'storage-volume-uris')
        if parent_sv_uris:
            try:
                parent_sv_uris.remove(self._uri)
            except ValueError:
                pass

        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties, email_to_addresses=None,
                          email_cc_addresses=None, email_insert=None):
        """
        Update writeable properties of this storage volume on the HMC, and
        optionally send emails to storage administrators requesting
        modification of the storage volume on the storage subsystem and of any
        resources related to the storage volume.

        This method performs the "Modify Storage Group Properties" operation,
        requesting modification of the volume.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the storage group owning this storage
          volume.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          properties (dict): New property values for the volume.
            Allowable properties are the fields defined in the
            "storage-volume-request-info" nested object for the "modify"
            operation. That nested object is described in section "Request body
            contents" for operation "Modify Storage Group Properties" in the
            :term:`HMC API` book.

            The properties provided in this parameter will be copied and then
            amended with the `operation="modify"` and `element-uri` properties,
            and then used as a single array item for the `storage-volumes`
            field in the request body of the "Modify Storage Group Properties"
            operation.

          email_to_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be notified.
            If `None` or empty, no email will be sent.

          email_cc_addresses (:term:`iterable` of :term:`string`): Email
            addresses of one or more storage administrator to be copied
            on the notification email.
            If `None` or empty, nobody will be copied on the email.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

          email_insert (:term:`string`): Additional text to be inserted in the
            notification email.
            The text can include HTML formatting tags.
            If `None`, no additional text will be inserted.
            Must be `None` or empty if `email_to_addresses` is `None` or empty.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        volreq_obj = copy.deepcopy(properties)
        volreq_obj['operation'] = 'modify'
        volreq_obj['element-uri'] = self.uri
        body = {
            'storage-volumes': [volreq_obj],
        }
        if email_to_addresses:
            body['email-to-addresses'] = email_to_addresses
            if email_cc_addresses:
                body['email-cc-addresses'] = email_cc_addresses
            if email_insert:
                body['email-insert'] = email_insert
        else:
            if email_cc_addresses:
                raise ValueError(
                    "email_cc_addresses must not be specified if there is no "
                    f"email_to_addresses: {email_cc_addresses!r}")
            if email_insert:
                raise ValueError(
                    "email_insert must not be specified if there is no "
                    f"email_to_addresses: {email_insert!r}")

        self.manager.session.post(
            self.manager.storage_group.uri + '/operations/modify',
            resource=self, body=body)

        self.update_properties_local(copy.deepcopy(properties))

    @logged_api_call
    def indicate_fulfillment_ficon(self, control_unit, unit_address):
        """
        TODO: Add ControlUnit objects etc for FICON support.

        Indicate completion of :term:`fulfillment` for this ECKD (=FICON)
        storage volume and provide identifying information (control unit
        and unit address) about the actual storage volume on the storage
        subsystem.

        Manually indicating fulfillment is required for all ECKD volumes,
        because they are not auto-discovered by the CPC.

        This method performs the "Fulfill FICON Storage Volume" HMC operation.

        Upon successful completion of this operation, the "fulfillment-state"
        property of this storage volume object will have been set to
        "complete". That is necessary for the CPC to be able to address and
        connect to the volume.

        If the "fulfillment-state" properties of all storage volumes in the
        owning storage group are "complete", the owning storage group's
        "fulfillment-state" property will also be set to "complete".

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the storage group owning this storage
          volume.
        * Task permission to the "Configure Storage - Storage Administrator"
          task.

        Parameters:

          control_unit (:class:`~zhmcclient.ControlUnit`):
            Logical control unit (LCU) in which the backing ECKD volume is
            defined.

          unit_address (:term:`string`):
            Unit address of the backing ECKD volume within its logical control
            unit,
            as a hexadecimal number of up to 2 characters in any lexical case.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        # The operation requires exactly 2 characters in lower case
        unit_address_2 = format(int(unit_address, 16), '02x')

        body = {
            'control-unit-uri': control_unit.uri,
            'unit-address': unit_address_2,
        }
        self.manager.session.post(
            self.uri + '/operations/fulfill-ficon-storage-volume',
            resource=self, body=body)

    @logged_api_call
    def indicate_fulfillment_fcp(self, wwpn, lun, host_port):
        """
        Indicate completion of :term:`fulfillment` for this FCP storage volume
        and provide identifying information (WWPN and LUN) about the actual
        storage volume on the storage subsystem.

        Manually indicating fulfillment is required for storage volumes that
        will be used as boot devices for a partition. The specified host
        port will be used to access the storage volume during boot of the
        partition.

        Because the CPC discovers storage volumes automatically, the
        fulfillment of non-boot volumes does not need to be manually indicated
        using this function (it may be indicated though before the CPC
        discovers a working communications path to the volume, but the role
        of the specified host port is not clear in this case).

        This method performs the "Fulfill FCP Storage Volume" HMC operation.

        Upon successful completion of this operation, the "fulfillment-state"
        property of this storage volume object will have been set to
        "complete". That is necessary for the CPC to be able to address and
        connect to the volume.

        If the "fulfillment-state" properties of all storage volumes in the
        owning storage group are "complete", the owning storage group's
        "fulfillment-state" property will also be set to "complete".

        HMC/SE version requirements:

        * :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the storage group owning this storage
          volume.
        * Task permission to the "Configure Storage - Storage Administrator"
          task.

        Parameters:

          wwpn (:term:`string`):
            World wide port name (WWPN) of the FCP storage subsystem containing
            the storage volume,
            as a hexadecimal number of up to 16 characters in any lexical case.

          lun (:term:`string`):
            Logical Unit Number (LUN) of the storage volume within its FCP
            storage subsystem,
            as a hexadecimal number of up to 16 characters in any lexical case.

          host_port (:class:`~zhmcclient.Port`):
            Storage port on the CPC that will be used to boot from.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        # The operation requires exactly 16 characters in lower case
        wwpn_16 = format(int(wwpn, 16), '016x')
        lun_16 = format(int(lun, 16), '016x')

        body = {
            'world-wide-port-name': wwpn_16,
            'logical-unit-number': lun_16,
            'adapter-port-uri': host_port.uri,
        }
        self.manager.session.post(
            self.uri + '/operations/fulfill-fcp-storage-volume',
            resource=self, body=body)



#!/usr/bin/env python
# Copyright 2020,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that discovers a storage group and prints the connection report
(DPM mode, z14).
"""

import sys
import json
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding CPCs in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)

    print("Selecting a z14 or higher CPC ...")
    cpc = None
    for _cpc in cpcs:
        se_version_info = [int(v)
                           for v in _cpc.get_property('se-version').split('.')]
        if se_version_info >= [2, 14]:
            cpc = _cpc
            break
    if not cpc:
        print(f"Error: HMC at {host} does not manage any z14 or higher CPC "
              "in DPM mode")
        sys.exit(1)
    se_version = cpc.get_property('se-version')
    print(f"Using CPC {cpc.name} (SE version: {se_version})")

    print(f"Listing storage groups of CPC {cpc.name} and selecting the "
          "first FCP storage group ...")
    storage_groups = cpc.list_associated_storage_groups()
    sg = None
    for _sg in storage_groups:
        if _sg.get_property('type') == 'fcp':
            sg = _sg
            break
    if not sg:
        print(f"Could not find any FCP storage group for CPC {cpc.name}")
        sys.exit(1)
    print(f"Using FCP storage group: {sg.name} (type: {sg.get_property('type')}, "
          f"shared: {sg.get_property('shared')}, "
          f"fulfillment: {sg.get_property('fulfillment-state')})")

    print(f"Listing partitions attached to storage group {sg.name} ...")
    parts = sg.list_attached_partitions()
    part_names = [p.name for p in parts]
    part_names_str = ', '.join(part_names) if part_names else "<none>"
    print(f"Partitions attached to storage group {sg.name}: {part_names_str}")

    print(f"Getting connection report for storage group {sg.name} ...")
    report = sg.get_connection_report()

    print("fcp-storage-subsystems section of connection report, before "
          "discovery:")
    print(json.dumps(report['fcp-storage-subsystems'], indent=2))

    print("Discovering LUNs of storage group (waiting for completion) ...")
    sg.discover_fcp()

    print(f"Getting connection report for storage group {sg.name} ...")
    report = sg.get_connection_report()

    print("fcp-storage-subsystems section of connection report, after "
          "discovery:")
    print(json.dumps(report['fcp-storage-subsystems'], indent=2))

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for activation profiles (with CPCs in classic mode).

These tests do not change any activation profiles.
"""


import warnings

import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import classic_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, runtest_find_list, \
    runtest_get_properties, setup_logging, End2endTestWarning, \
    skip_missing_api_feature

urllib3.disable_warnings()

# Logging for zhmcclient HMC interactions and test functions
LOGGING = False
LOG_FILE = 'test_profile.log'

# Properties in minimalistic ActivationProfile objects (e.g. find_by_name())
ACTPROF_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in ActivationProfile objects returned by list() without full props
ACTPROF_LIST_PROPS = ['element-uri', 'name', 'target-name']

# Properties in image ActivationProfile objects for list(additional_properties)
ACTPROF_ADDITIONAL_PROPS = ['description', 'ipl-address']

# Properties whose values can change between retrievals
ACTPROF_VOLATILE_PROPS = []


def standard_activation_profile_props(cpc, profile_name, profile_type):
    """
    Return the input properties for creating standard activation profile in the
    specified CPC.
    """
    actprof_input_props = {
        'profile-name': profile_name,
        'description': (
            f'{profile_type} profile for zhmcclient end2end tests'),
    }
    if profile_type == 'image':
        # We provide the minimum set of properties needed to create a profile.
        if cpc.prop('processor-count-ifl'):
            actprof_input_props['number-shared-ifl-processors'] = 1
            actprof_input_props['operating-mode'] = 'linux-only'
        elif cpc.prop('processor-count-general-purpose'):
            actprof_input_props['number-shared-general-purpose-processors'] = 1
            actprof_input_props['operating-mode'] = 'esa390'
        else:
            actprof_input_props['number-shared-general-purpose-processors'] = 1
            actprof_input_props['operating-mode'] = 'esa390'
            warnings.warn(
                f"CPC {cpc.name} shows neither IFL nor CP processors, "
                "specifying 1 CP for image activation profile creation.",
                End2endTestWarning)

    return actprof_input_props


@pytest.mark.parametrize(
    "profile_type", ['reset', 'image', 'load']
)
def test_actprof_crud(classic_mode_cpcs, profile_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete an activation profile.
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    logger = setup_logging(LOGGING, 'test_actprof_crud', LOG_FILE)

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        console = cpc.manager.console
        skip_missing_api_feature(
            console, 'create-delete-activation-profiles',
            cpc, 'create-delete-activation-profiles')

        actprof_mgr = getattr(cpc, profile_type + '_activation_profiles')

        msg = f"Testing on CPC {cpc.name}"
        print(msg)
        logger.info(msg)

        actprof_name = f'ZHMC{profile_type[0].upper()}1'

        # Preparation: Ensure clean starting point for this test
        try:
            _actprof = actprof_mgr.find(name=actprof_name)
        except zhmcclient.NotFound:
            pass
        else:
            msg = (
                f"Preparation: Delete {profile_type} activation profile "
                f"{actprof_name!r} on CPC {cpc.name} from previous run")
            warnings.warn(msg, UserWarning)
            logger.info(msg)
            _actprof.delete()

        # Test creating the activation profile
        actprof_input_props = standard_activation_profile_props(
            cpc, actprof_name, profile_type)

        logger.info("Test: Create %s activation profile %r on CPC %s",
                    profile_type, actprof_name, cpc.name)

        # The code to be tested
        actprof = actprof_mgr.create(actprof_input_props)

        try:
            for pn, exp_value in actprof_input_props.items():
                assert actprof.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            actprof.pull_full_properties()
            for pn, exp_value in actprof_input_props.items():
                if pn == 'profile-name':
                    pn = 'name'
                assert actprof.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"

            # Test updating a property of the activation profile

            new_desc = "Updated activation profile description."

            logger.info("Test: Update a property of %s activation profile "
                        "%r on CPC %s", profile_type, actprof_name, cpc.name)

            # The code to be tested
            actprof.update_properties(dict(description=new_desc))

            assert actprof.properties['description'] == new_desc
            actprof.pull_full_properties()
            assert actprof.properties['description'] == new_desc

        finally:
            # Test deleting the activation profile (also cleanup)

            logger.info("Test: Delete %s activation profile %r on CPC %s",
                        profile_type, actprof_name, cpc.name)

            # The code to be tested
            actprof.delete()


@pytest.mark.parametrize(
    "profile_type", ['reset', 'image', 'load']
)
def test_actprof_find_list(classic_mode_cpcs, profile_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition
        actprof_mgr = getattr(cpc, profile_type + '_activation_profiles')

        # Pick the activation profiles to test with
        actprof_list = actprof_mgr.list()
        if not actprof_list:
            skip_warn(
                f"No {profile_type} activation profiles on CPC {cpc.name} "
                f"managed by HMC {hd.host}")
        actprof_list = pick_test_resources(actprof_list)

        for actprof in actprof_list:
            print(f"Testing on CPC {cpc.name} with {profile_type} activation "
                  f"profile {actprof.name!r}")
            if profile_type == 'image':
                actprof_additional_props = ACTPROF_ADDITIONAL_PROPS
            else:
                actprof_additional_props = None
            runtest_find_list(
                session, actprof_mgr, actprof.name, 'name', 'element-uri',
                ACTPROF_VOLATILE_PROPS, ACTPROF_MINIMAL_PROPS,
                ACTPROF_LIST_PROPS, actprof_additional_props)


@pytest.mark.parametrize(
    "profile_type", ['reset', 'image', 'load']
)
def test_actprof_property(classic_mode_cpcs, profile_type):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not classic_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in classic mode")

    for cpc in classic_mode_cpcs:
        assert not cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition
        actprof_mgr = getattr(cpc, profile_type + '_activation_profiles')

        # Pick the activation profiles to test with
        actprof_list = actprof_mgr.list()
        if not actprof_list:
            skip_warn(
                f"No {profile_type} activation profiles on CPC {cpc.name} "
                f"managed by HMC {hd.host}")
        actprof_list = pick_test_resources(actprof_list)

        for actprof in actprof_list:
            print(f"Testing on CPC {cpc.name} with {profile_type} activation "
                  f"profile {actprof.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(actprof.manager, non_list_prop)



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pytest fixtures for CPCs.
"""


import warnings
import pytest
import zhmcclient

# pylint: disable=unused-import
from ._hmc_definition_fixtures import hmc_session  # noqa: F401

__all__ = ['all_cpcs', 'dpm_mode_cpcs', 'classic_mode_cpcs']


def fixtureid_cpcs(fixture_value):
    """
    Return a fixture ID to be used by pytest, for fixtures returning a list of
    CPCs.

    Parameters:

      * fixture_value (list of zhmcclient.Cpc): The list of CPCs.
    """
    if not isinstance(fixture_value, list):
        return None  # Use pytest auto-generated ID
    cpc_list = fixture_value
    cpcs_str = ','.join([cpc.name for cpc in cpc_list])
    return f"CPCs={cpcs_str}"


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def all_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of all CPCs to test against, regardless
    of their operational mode.

    The CPCs to test against are defined in the ``cpcs`` variable for the
    HMC entry in the :ref:`HMC inventory file`.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'any')


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def dpm_mode_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of CPCs in DPM mode to test against.

    The CPCs to test against are defined in the ``cpcs`` variable in the
    HMC entry in the :ref:`HMC inventory file` and have their ``dpm_enabled``
    property set to ``true``.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'dpm')


@pytest.fixture(
    scope='module',
    ids=fixtureid_cpcs
)
def classic_mode_cpcs(request, hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of CPCs in classic mode to test against.

    The CPCs to test against are defined in the ``cpcs`` variable in the
    HMC entry in the :ref:`HMC inventory file` and have their ``dpm_enabled``
    property set to ``false``.

    A test function parameter with the name of this fixture resolves to a list
    of :class:`zhmcclient.Cpc` objects representing that set of CPCs. These
    objects have the "short" set of properties from :meth:`zhmcclient.Cpc.list`.

    The test function is invoked just once with the list of CPCs, and the
    test function needs to loop through the CPCs (or a subset).

    Because the `hmc_session` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_session`
    function needs to be imported as well when this fixture is used.
    """
    return defined_cpcs(hmc_session, 'classic')


def defined_cpcs(session, mode):
    """
    Return a list of CPCs defined in the HMC definition file, that are actually
    managed by the HMC, and that actually have the desired operational mode.

    Parameters:

      session (:class:`zhmcclient.Session`): The session with the HMC.

      mode (string): The desired mode of the CPC ('dpm', 'classic', 'any').

    Returns:

      list of :class:`zhmcclient.Cpc`: The CPCs in the desired mode.
    """
    client = zhmcclient.Client(session)
    actual_cpcs = client.cpcs.list()
    actual_cpc_names = [cpc.name for cpc in actual_cpcs]
    hd = session.hmc_definition
    result_cpcs = []
    for cpc_name in hd.cpcs:
        cpcs = [cpc for cpc in actual_cpcs if cpc.name == cpc_name]
        if not cpcs:
            msg_txt = (
                f"CPC {cpc_name} defined for HMC {hd.nickname} at {hd.host} "
                "in HMC definition file is not managed by that HMC. "
                f"Actually managed CPCs: {', '.join(actual_cpc_names)}")
            warnings.warn(msg_txt, UserWarning)
            pytest.skip(msg_txt)
        cpc = cpcs[0]

        if mode == 'dpm' and not cpc.dpm_enabled:
            continue
        if mode == 'classic' and cpc.dpm_enabled:
            continue
        result_cpcs.append(cpc)

    return result_cpcs



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Encapsulation of an HMC inventory file in YAML format.

HMC inventory files conform to the format of HMC inventory files in YAML
format and define specific additional variables for HMCs.
"""


from collections import OrderedDict
import errno
import yaml
import yamlloader
import jsonschema

__all__ = ['HMCInventoryFileError', 'HMCInventoryFile']

# Structure of an HMC inventory file in YAML format:
#
#   all:  # the top-level group
#     hosts:
#       <hmc_name>:  # DNS hostname, IP address, or nickname of HMC
#         description: <string>
#         contact: <string>
#         access_via: <string>
#         ansible_host: <host> or [<host>, ...]   # if real HMC
#         mock_file: <path_name>                  # if mocked HMC
#         cpcs:
#           <cpc_name>:
#             <prop_name>: <prop_value>
#         <var_name>: <var_value>  # additional variables for host
#     vars:
#       <var_name>: <var_value>  # additional variables for all hosts in group
#     children:
#       <group_name>:  # a child group
#         hosts: ...
#         vars: ...
#         children: ...

# Regexp for valid Ansible variable names, host aliases, group names
ANSIBLE_NAME_PATTERN = r"^[a-zA-Z_][a-zA-Z_0-9]*$"

# Valid Ansible variable types (using JSON schema type names)
ANSIBLE_VARTYPES = [
    "object", "array", "string", "integer", "number", "boolean", "null"
]

# Regexp for valid Ansible hosts (DNS name, IP address, or host alias)
# Note: Host name ranges like in Ansible are not supported.
ANSIBLE_HOST_PATTERN = r"^[a-zA-Z_0-9:\.\-]+$"

# Regexp for valid HMC CPC names
CPCNAME_PATTERN = r"^[A-Z][A-Z0-9_]*$"

# Regexp for valid HMC resource property names (with underscores instead of
# hyphens)
PROPNAME_PATTERN = r"^[a-z][a-z0-9_]*$"

# JSON schema for content of an HMC inventory file in YAML format
HMC_INVENTORY_FILE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "title": "JSON schema for an HMC inventory file in YAML format",
    "definitions": {
        "HostVars": {
            "description": "Variables for the host",
            "type": ["object", "null"],
            "additionalProperties": False,
            "properties": {
                "description": {
                    "description": "Short description of the HMC. "
                                   "Optional, default: empty.",
                    "type": "string",
                },
                "contact": {
                    "description": "Name of technical contact for the HMC. "
                                   "Optional, default: empty.",
                    "type": "string",
                },
                "access_via": {
                    "description": "Preconditions to reach the network "
                                   "of the HMC. "
                                   "Optional, default: empty.",
                    "type": "string",
                },
                "ansible_host": {
                    "description": "For real HMCs: DNS host name or IP "
                                   "address of the HMC or of a list of "
                                   "redundant HMCs. "
                                   "Mandatory for real HMCs.",
                    "type": ["string", "array"],
                },
                "mock_file": {
                    "description": "For mocked HMCs: Path name of HMC mock "
                                   "file, relative to the directory of this "
                                   "HMC inventory file. "
                                   "Mandatory for mocked HMCs.",
                    "type": "string",
                },
                "cpcs": {
                    "description": "CPCs managed by this HMC that are to be "
                                   "tested. If omitted or None, no CPCs are "
                                   "tested.",
                    "type": "object",
                    "additionalProperties": False,
                    "patternProperties": {
                        # CPC name:
                        CPCNAME_PATTERN: {
                            "description":
                                "Expected CPC properties. Used for basic "
                                "classification of the CPC, e.g. 'dpm-mode', "
                                "'machine-type', 'machine-model'.",
                            "type": "object",
                            "additionalProperties": False,
                            "patternProperties": {
                                # CPC property name (with underscores):
                                PROPNAME_PATTERN: {
                                    "description": "Expected property value",
                                    "type": ANSIBLE_VARTYPES,
                                },
                            },
                        },
                    },
                },
            },
            "patternProperties": {
                # Additional variables for the host.
                # Variable name:
                ANSIBLE_NAME_PATTERN: {
                    "description": "Variable value",
                    "type": ANSIBLE_VARTYPES,
                },
            },
        },
        "GroupVars": {
            "description":
                "Additional variables for all hosts in the group",
            "type": ["object", "null"],
            "additionalProperties": False,
            "patternProperties": {
                # Variable name:
                ANSIBLE_NAME_PATTERN: {
                    "description": "Variable value",
                    "type": ANSIBLE_VARTYPES,
                },
            },
        },
        "Hosts": {
            "description": "Hosts in the group",
            "type": ["object", "null"],
            "additionalProperties": False,
            "patternProperties": {
                # Host DNS name, IP address, or alias:
                ANSIBLE_HOST_PATTERN: {
                    "$ref": "#/definitions/HostVars"
                },
            },
        },
        "Children": {
            "description": "Child groups of a group",
            "type": ["object", "null"],
            "additionalProperties": False,
            "patternProperties": {
                # Child group name:
                ANSIBLE_NAME_PATTERN: {
                    "description": "Content of the child group",
                    "type": ["object", "null"],
                    "additionalProperties": False,
                    "properties": {
                        "hosts": {
                            "$ref": "#/definitions/Hosts"
                        },
                        "vars": {
                            "$ref": "#/definitions/GroupVars"
                        },
                        "children": {
                            "$ref": "#/definitions/Children"
                        },
                    },
                },
            },
        },
    },
    "description": "Top-level group named 'all'",
    "type": "object",
    "additionalProperties": False,
    "required": [
        "all",
    ],
    "properties": {
        # Top-level group name:
        "all": {
            "description": "Content of the top-level group",
            "type": ["object", "null"],
            "additionalProperties": False,
            "properties": {
                "hosts": {
                    "$ref": "#/definitions/Hosts"
                },
                "vars": {
                    "$ref": "#/definitions/GroupVars"
                },
                "children": {
                    "$ref": "#/definitions/Children"
                },
            },
        },
    },
}


class HMCInventoryFileError(Exception):
    """
    An error in the :ref:`HMC inventory file`.
    """
    pass


class HMCInventoryFile:
    """
    Encapsulation of an :ref:`HMC inventory file` in YAML format.
    """

    def __init__(self, filepath):
        self._filepath = filepath
        self._data = OrderedDict()  # File content
        self._load_file()

    def _load_file(self):
        """
        Load and validate the :ref:`HMC inventory file` in YAML format.
        """
        try:
            # pylint: disable=unspecified-encoding
            with open(self._filepath) as fp:
                try:
                    data = yaml.load(
                        fp, Loader=yamlloader.ordereddict.SafeLoader)
                except (yaml.parser.ParserError,
                        yaml.scanner.ScannerError) as exc:
                    new_exc = HMCInventoryFileError(
                        "Invalid YAML syntax in HMC inventory file "
                        f"{self._filepath!r}: {exc.__class__.__name__} {exc}")
                    new_exc.__cause__ = None
                    raise new_exc  # HMCInventoryFileError
        except OSError as exc:
            if exc.errno == errno.ENOENT:
                new_exc = HMCInventoryFileError(
                    f"The HMC inventory file {self._filepath!r} was not found")
                new_exc.__cause__ = None
                raise new_exc  # HMCInventoryFileError
            raise

        try:
            jsonschema.validate(data, HMC_INVENTORY_FILE_SCHEMA)
        except jsonschema.exceptions.ValidationError as exc:
            elem = '.'.join(str(e) for e in exc.absolute_path)
            schemaitem = '.'.join(str(e) for e in exc.absolute_schema_path)
            new_exc = HMCInventoryFileError(
                "Invalid data format in HMC inventory file "
                f"{self._filepath}: {exc.message}; "
                f"Offending element: {elem}; "
                f"Schema item: {schemaitem}; "
                f"Validator: {exc.validator}={exc.validator_value}")
            new_exc.__cause__ = None
            raise new_exc  # HMCInventoryFileError

        self._data.update(data)

    def __repr__(self):
        return (
            "HMCVaultFile("
            f"filepath={self.filepath!r}, "
            f"data={self.data!r})")

    @property
    def filepath(self):
        """
        string: Path name of the :ref:`HMC inventory file`.
        """
        return self._filepath

    @property
    def data(self):
        """
        :class:`~py:collections.OrderedDict`: Content of the
        :ref:`HMC inventory file`, as nested
        :class:`~py:collections.OrderedDict` and
        :class:`~py:list` objects.
        """
        return self._data



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that performs an asynchronous start of a partition and waits for a job
completion notification.
"""

import sys
import uuid
import requests.packages.urllib3
import stomp
import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

# HMC port for JMS notifications
JMS_PORT = 61612

# Notification topic type for jobs
JOB_TOPIC_TYPE = 'job-notification'

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding job completion notification topic ...")
    job_topic_name = None
    topics = session.get_notification_topics()
    for topic in topics:
        if topic['topic-type'] == JOB_TOPIC_TYPE:
            job_topic_name = topic['topic-name']
            break
    print(f"Using job completion notification topic: {job_topic_name}")

    print("Finding a CPC in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    part_name = f"zhmc_test_{uuid.uuid4()}"
    print(f"Creating partition {part_name} ...")
    try:
        part = cpc.partitions.create(
            properties={
                'name': part_name,
                'type': 'linux',
                'ifl-processors': 2,
                'initial-memory': 4096,
                'maximum-memory': 4096,
            })
    except zhmcclient.Error as exc:
        print(f"Error: Cannot create partition {part_name} on CPC {cpc.name}: "
              f"{exc.__class__.__name__}: {exc}")
        sys.exit(1)

    try:
        print(f"Starting partition {part.name} asynchronously ...")
        job = part.start(wait_for_completion=False)

        print("Creating a notification receiver for topic "
              f"{job_topic_name} ...")
        try:
            receiver = zhmcclient.NotificationReceiver(
                job_topic_name, host, userid, password)
        except Exception as exc:
            print(f"Error: Cannot create notification receiver: {exc}")
            sys.exit(1)

        print("Waiting for job completion notifications ...")
        while True:
            try:
                for headers, _ in receiver.notifications():
                    # message is None for job completion notifications
                    if headers['job-uri'] == job.uri:
                        print("Received completion notification for the start job")
                        break
                    else:
                        print("Received completion notification for another job: "
                              f"{headers['job-uri']} - continue to wait")
            except zhmcclient.NotificationError as exc:
                print(f"Notification Error: {exc} - reconnecting")
                continue
            except stomp.exception.StompException as exc:
                print(f"STOMP Error: {exc} - reconnecting")
                continue
            except KeyboardInterrupt:
                print("Keyboard interrupt - leaving receiver loop")
                receiver.close()
                break
            else:
                raise AssertionError("Receiver was closed - should not happen")

        print("Job completed; Getting job status and result ...")
        try:
            job_status, op_result = job.check_for_completion()
        except zhmcclient.Error as exc:
            print("Error: Start operation failed with "
                  f"{exc.__class__.__name__}: {exc}")
            sys.exit(1)
        assert job_status == 'complete'
        print("Start operation succeeded")

    finally:
        if part.get_property('status') != 'stopped':
            print(f"Stopping partition {part.name} ...")
            try:
                part.stop(wait_for_completion=True)
            except zhmcclient.Error as exc:
                print("Error: Stop operation failed with "
                      f"{exc.__class__.__name__}: {exc}")
                sys.exit(1)

        print(f"Deleting partition {part.name} ...")
        try:
            part.delete()
        except zhmcclient.Error as exc:
            print(f"Error: Cannot delete partition {part.name} on CPC "
                  f"{cpc.name} for clean up - "
                  f"Please delete it manually: {exc.__class__.__name_}: {exc}")
            sys.exit(1)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for testutils._hmc_inventory_file module.
"""


import os
import tempfile
import pytest

from zhmcclient.testutils._hmc_inventory_file import HMCInventoryFile, \
    HMCInventoryFileError


# Test cases for HMCInventoryFile.__init__()
TESTCASES_HMC_INVENTORY_FILE_INIT = [
    # Format:
    # * desc: Testcase description
    # * file_data: String with data of the HMC inventory file to test
    # * exp_exc_type: Expected exception type, or None for success
    (
        "Typical inventory file",
        """
all:
  hosts:
    globalservices:  # host alias
      description: bla
      contact: bla
      access_via: bla
      ansible_host: 1.2.3.4
      cpcs:
        CPC1:
          dpm_mode: true
      extra_host_var1: bla
    global.services.com:  # host DNS name
      description: bla
      contact: bla
      access_via: bla
      cpcs:
        CPC1:
          dpm_mode: true
      extra_host_var1: bla
    1.2.3.4:  # IPv4 address
      extra_host_var1: bla
    "01:02:03:04:0a::0d":  # IPv6 address
      extra_host_var1: bla
  vars:
    extra_group_var1: bla
  children:
    east:
      hosts:
        1.2.3.5:
          description: bla
      vars:
        extra_group_var1: bla
      children:
        dbserver:
          hosts:
            1.2.3.6:
              extra_group_var1: bla
        """,
        None
    ),
    (
        "DNS hostname without variables",
        """
all:
  hosts:
    app.services.com:
        """,
        None
    ),
    (
        "DNS hostname with numeric range - not supported",
        """
all:
  hosts:
    app[1:20:2].services.com:
        """,
        HMCInventoryFileError
    ),
    (
        "DNS hostname with alphabetic range - not supported",
        """
all:
  hosts:
    app[a:f].services.com:
        """,
        HMCInventoryFileError
    ),
    (
        "Top-level group with empty hosts, vars, children",
        """
all:
  hosts:
  vars:
  children:
        """,
        None
    ),
    (
        "Empty top-level group",
        """
all:
        """,
        None
    ),
    (
        "Schema validation error: Required top-level element missing",
        """
xx:
        """,
        HMCInventoryFileError
    ),
    (
        "Schema validation error: Invalid additional element",
        """
all:
  xx:
        """,
        HMCInventoryFileError
    ),
    (
        "Schema validation error: List instead of dict",
        """
all: []
        """,
        HMCInventoryFileError
    ),
    (
        "YAML scanner error",
        """
all: xx: yy:
        """,
        HMCInventoryFileError
    ),
    (
        "YAML parser error",
        """
 all: x
xx: x
        """,
        HMCInventoryFileError
    ),
]


@pytest.mark.parametrize(
    "desc, file_data, exp_exc_type",
    TESTCASES_HMC_INVENTORY_FILE_INIT)
def test_HMCInventoryFile_init(desc, file_data, exp_exc_type):
    # pylint: disable=unused-argument,invalid-name
    """
    Test function for HMCInventoryFile.__init__().
    """

    with tempfile.NamedTemporaryFile(
            mode='w', suffix='.yaml', delete=False) as tf:
        tf.write(file_data)

    try:
        if exp_exc_type:
            with pytest.raises(exp_exc_type) as exc_info:

                # The function to be tested
                HMCInventoryFile(tf.name)

            exc = exc_info.value
            assert isinstance(exc, exp_exc_type)

        else:

            # The function to be tested
            _ = HMCInventoryFile(tf.name)

    finally:
        os.remove(tf.name)



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that finds an LPAR on a CPC in classic mode and performs
activate/load/deactivate on the LPAR.
"""

import sys
import time
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

# Customize: Set to the LPAR that you want to activate/load/deactivate
lpar_name = 'ZHMCTEST'

# Customize: Set to the device number from which the LPAR should load
lpar_load_devno = '0100'

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in classic mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': False})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in classic mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    print(f"Finding LPAR by name={lpar_name} ...")
    # We use list() instead of find() because find(name=..) is optimized by
    # using the name-to-uri cache and therefore returns an Lpar object with
    # only a minimal set of properties, and particularly no 'status' property.
    # That would drive an extra "Get Logical Partition Properties" operation
    # when the status property is accessed.
    lpars = cpc.lpars.list(filter_args={'name': lpar_name})
    if len(lpars) != 1:
        print(f"Error: Could not find LPAR {lpar_name} on CPC {cpc.name} - "
              "customize the LPAR name in the example script")
        lpar_names = [lpar.name for lpar in cpc.lpars.list()]
        lpar_str = ', '.join(lpar_names)
        print(f"Note: The following LPARs exist on CPC {cpc.name}: {lpar_str}")
        sys.exit(1)
    lpar = lpars[0]
    status = lpar.get_property('status')
    print(f"Found LPAR {lpar.name} with status {status}")

    retries = 10

    if status != "not-activated":
        print(f"Deactivating LPAR {lpar.name} ...")
        lpar.deactivate()
        for i in range(0, retries):
            lpar = cpc.lpars.list(filter_args={'name': lpar_name})[0]
            status = lpar.get_property('status')
            print(f"LPAR status: {status}")
            if status == 'not-activated':
                break
            time.sleep(1)
        else:
            print(f"Warning: After {retries} retries, status of LPAR "
                  f"{lpar.name} after Deactivate is still: {status}")

    print(f"Activating LPAR {lpar.name} ...")
    lpar.activate()
    for i in range(0, retries):
        lpar = cpc.lpars.list(filter_args={'name': lpar_name})[0]
        status = lpar.get_property('status')
        print(f"LPAR status: {status}")
        if status == 'not-operating':
            break
        time.sleep(1)
    else:
        print(f"Warning: After {retries} retries, status of LPAR {lpar.name} "
              f"after Activate is still: {status}")

    print(f"Loading LPAR {lpar.name} from device {lpar_load_devno} ...")
    lpar.load(lpar_load_devno)
    for i in range(0, retries):
        lpar = cpc.lpars.list(filter_args={'name': lpar_name})[0]
        status = lpar.get_property('status')
        print(f"LPAR status: {status}")
        if status == 'operating':
            break
        time.sleep(1)
    else:
        print(f"Warning: After {retries} retries, status of LPAR {lpar.name} "
              f"after Load is still: {status}")

    print(f"Deactivating LPAR {lpar.name} ...")
    lpar.deactivate()
    for i in range(0, retries):
        lpar = cpc.lpars.list(filter_args={'name': lpar_name})[0]
        status = lpar.get_property('status')
        print(f"LPAR status: {status}")
        if status == 'not-activated':
            break
        time.sleep(1)
    else:
        print(f"Warning: After {retries} retries, status of LPAR {lpar.name} "
              f"after Deactivate is still: {status}")

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for capacity groups (on CPCs in DPM mode).

These tests do not change any existing capacity groups, but create, modify and
delete test capacity groups.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic CapacityGroup objects (e.g. find_by_name())
CAPGRP_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in CapacityGroup objects returned by list() without full props
CAPGRP_LIST_PROPS = ['element-uri', 'name']

# Properties whose values can change between retrievals of CapacityGroup objects
CAPGRP_VOLATILE_PROPS = []


def test_capgrp_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the capacity groups to test with
        capgrp_list = cpc.capacity_groups.list()
        if not capgrp_list:
            skip_warn(f"No capacity groups defined on CPC {cpc.name} managed "
                      f"by HMC {hd.host}")
        capgrp_list = pick_test_resources(capgrp_list)

        for capgrp in capgrp_list:
            print(f"Testing on CPC {cpc.name} with capacity group "
                  f"{capgrp.name!r}")
            runtest_find_list(
                session, cpc.capacity_groups, capgrp.name, 'name',
                'element-uri', CAPGRP_VOLATILE_PROPS, CAPGRP_MINIMAL_PROPS,
                CAPGRP_LIST_PROPS)


def test_capgrp_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the capacity groups to test with
        capgrp_list = cpc.capacity_groups.list()
        if not capgrp_list:
            skip_warn(f"No capacity groups defined on CPC {cpc.name} managed "
                      f"by HMC {hd.host}")
        capgrp_list = pick_test_resources(capgrp_list)

        for capgrp in capgrp_list:
            print(f"Testing on CPC {cpc.name} with capacity group "
                  f"{capgrp.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(capgrp.manager, non_list_prop)


def test_capgrp_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a capacity group.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        print(f"Testing on CPC {cpc.name}")

        capgrp_name = TEST_PREFIX + ' test_capgrp_crud capgrp1'
        capgrp_name_new = capgrp_name + ' new'

        # Ensure a clean starting point for this test
        try:
            capgrp = cpc.capacity_groups.find(name=capgrp_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test capacity group from previous run: "
                f"{capgrp_name!r} on CPC {cpc.name}", UserWarning)
            capgrp.delete()
        try:
            capgrp = cpc.capacity_groups.find(name=capgrp_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test capacity group from previous run: "
                f"{capgrp_name_new!r} on CPC {cpc.name}", UserWarning)
            capgrp.delete()

        # Test creating the capacity group

        capgrp_input_props = {
            'name': capgrp_name,
            'description': 'Test capacity group for zhmcclient end2end tests',
            'absolute-ifl-proc-cap': 1.0,
        }
        capgrp_auto_props = {
            'capping-enabled': True,
        }

        # The code to be tested
        capgrp = cpc.capacity_groups.create(capgrp_input_props)

        for pn, exp_value in capgrp_input_props.items():
            assert capgrp.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        capgrp.pull_full_properties()
        for pn, exp_value in capgrp_input_props.items():
            assert capgrp.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        for pn, exp_value in capgrp_auto_props.items():
            assert capgrp.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"

        # Test updating a property of the capacity group

        new_desc = "Updated capacity group description."

        # The code to be tested
        capgrp.update_properties(dict(description=new_desc))

        assert capgrp.properties['description'] == new_desc
        capgrp.pull_full_properties()
        assert capgrp.properties['description'] == new_desc

        # Test renaming the capacity group

        # The code to be tested
        capgrp.update_properties(dict(name=capgrp_name_new))

        assert capgrp.properties['name'] == capgrp_name_new
        capgrp.pull_full_properties()
        assert capgrp.properties['name'] == capgrp_name_new
        with pytest.raises(zhmcclient.NotFound):
            cpc.capacity_groups.find(name=capgrp_name)

        # Test deleting the capacity group

        # The code to be tested
        capgrp.delete()

        with pytest.raises(zhmcclient.NotFound):
            cpc.capacity_groups.find(name=capgrp_name_new)
